//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BJS_IC.h,v 1.10 2003/01/06 13:50:16 bouchez Exp $
//
// Description : Prototype de la classe BJS_IC.
// Divers : 
//-----------------------------------------------------------------------------

#ifndef _BJS_IC_H
#define _BJS_IC_H

#include "BioJeuSingle.h"

/** Classe des jeux de donnes intercross. */
class BJS_IC : public BioJeuSingle {
 public:
  
  /** Constructeur. */
  BJS_IC();
  
  /** Constructeur.
      @param cartag accs  l'ensemble des informations du systme. 
      @param id le numro du jeu.
      @param cross type de jeu de donne.
      @param nm nombre de marqueurs.
      @param taille de l'chantillon
      @param bitjeu champ de bit du jeu de donnes.
      @param indmarq le vecteur d'indice de marqueur
      @param echantil la matrice des informations
  */
  BJS_IC(CartaGenePtr cartag,
	 int id,
	 charPtr nomjeu,
	 CrossType cross,
	 int nm, 
	 int taille,
	 int bitjeu,
	 int *indmarq,
	 Obs **echantil);
  
  /** Destructeur. */
  ~BJS_IC();
  
  /** Compatibilite entre deux echantillons d'un marqueur
      @param numarq1 le numero du marqueur 1
      @param numarq2 le numero du marqueur 2
      @param numind le numero de l'individu
  */
  int Compatible(int numarq1, int numarq2) const;



  /** Fusion de deux marqueurs pour la creation d'un haplotype. Le
      marqueur 1 est utilise pour stocker le resultat.
      @param numarq1 marqueur 1 
      @param numarq2 marqueur 2 
      @return nothing.  
  */
  void Merge(int marq1,int marq2) const;


  /** Calcul du log sous l'hypothse d'indpendance 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param nbdata le nombre de donnes informatives(o)
      @return le log
  */
  double LogInd(int m1, 
		int m2,
		int *nbdata) const;
  
  /** Calcul l'esprance du nombre de recombinants, (log volutif).
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param theta
      @param loglike(o)
      @return l'esprance      
  */
  double EspRec(int m1, 
		int m2, 
		double theta, 
		double *loglike) const;

  /** Etape d'Expectation de EM.
      @param data la carte
      @param expected le vecteur d'expectation
      @return le loglike
  */
  double ComputeExpected(const Carte *data, double *expected);
  
 private:
    
  /** Prparation des structures pour la programmation dynamique de EM.
      @param data la carte
  */
  void PreparEM(const Carte *data);
  
  /** nettoyage des structures pour la programmation dynamique de EM.
      @param data la carte
  */
  void NetEM(const Carte *data);


  /** Calcul stat. suffisantes pour le 2pt
      @param m1 premier marqueur
      @param m2 second marqueur
      @param ss statistiques suffisantes (n[1..3])
  */
  void Prepare2pt(int m1, int m2,  int *ss) const {
    sprintf(bouf, "Bug, mthode interdite = Prepare2pt!\n");
    pout(bouf);
    return;
  };

  /** Calcul des estimateurs 2pt de vrais. max
      @param par vecteur des parametres du modele (theta,r mis a jour)
      @param ss statistiques suffisantes (n[1..3])
  */
  void Estimate2pt(double *par, int *ss) const  {
    sprintf(bouf, "Bug, mthode interdite = Estimate2pt!\n");
    pout(bouf);
    return;
  };

  /** Calcul de la vraisemblance 2pt etant donnes les parametres du
      modele.
      @param par vecteur des parametres du modele (theta,r)
      @param ss statistiques suffisantes (n[1..3])
  */
  double LogLike2pt(double *par, int *ss) const  {
    sprintf(bouf, "Bug, mthode interdite = LogLike2pt!\n");
    pout(bouf);
    return 0.0;
  };

  /** Calcul du LOD score et de fraction de recombinants 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @param epsilon le seuil de convergence
      @param fr la fraction de recombinants(i,o)
      @return le LOD
  */
  double ComputeOneTwoPoints(int m1, 
			     int m2, 
			     double epsilon,
			     double *fr) const; 

  /** tableaux des probabilits conditionnelles de passage par un sommet depuis la gauche */
  doublePtr SourceTo[4];

  /** tableaux des probabilits conditionnelles de passage par un sommet depuis la droite */
  doublePtr ToSink[4];

  /** Calcul de la probabilite de passage par une arete suivant la paire de genotypes relies
      @param a le premier gnotype
      @param b le second gnotype
      @param p la proba de recombinaison
      @return la probabilit
  */
  double ProbArete(int a, int b, double p) const;

  /** Calcul des probabilites d''existence d''un chemin "depuis la gauche".
      @param Ind l'individu
      @param map la carte
      @return le loglike
    */
  double ComputeSourceTo(int Ind, const Carte *map) const;

  /** Calcul des probabilites d''existence d''un chemin "depuis la droite".
      @param Ind l'individu
      @param map la carte
      @return le loglike
  */
  double ComputeToSink(int Ind, const Carte *map) const;

  /** Etape elementaire d''Expectation de EM.
      @param Ind l'individu
      @param map la carte
      @param expected vecteur d'expectation
  */
  void ComputeOneExpected(int Ind, const Carte* map, double *expected) const;

};

#endif /* _BJS_IC_H */
