//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: Tas.h,v 1.23 2003/12/17 07:50:59 tschiex Exp $
//
// Description : Prototype de la classe Tas. Gestion de cartes.
// Divers :
//-----------------------------------------------------------------------------


#ifndef _TAS_H
#define _TAS_H

#include "CGtypes.h"
#include "Carte.h"
#include "System.h"

class CartaGene;

typedef CartaGene *CartaGenePtr;

#define HASHSIZE 2069
#define MAXHEAPSIZE 15
#define MAXHITS 3

/** Classe de stockage d'informations sur une carte */
class StructHMap {

 public:

  /** La carte stocke dans le tas */
  Carte *map;
  
  /** Temps qu'il a fallu pour trouver cette carte */
  double TimeStamp;

  /** */
  unsigned short BirthDate;

  /** Nombre de fois que la carte a t lue */
  unsigned short NbHits;
  
  /** Pointeur sur la carte suivante dans la liste */
  StructHMap *Next;
  
  /** */
  StructHMap **InH;

  /** Constructeur 
    @param cartage accs aux informations du systme
    @param n le nombre de marqueurs
  */
  StructHMap(CartaGene *cartage, int n);

  /** Destructeur */
  ~StructHMap();

};


/** Classe de gestion de cartes dans un tas */
class Tas {
 public:

  /** Accs  l'ensemble des informations du systme. */
  CartaGene *Cartage;

  /** Drapeau du mode empchant de stocker des cartes redondantes
   (cas de la fusion sur l'ordre)
  */
  int EquivalenceFlag;

  /** */
  int HeapSize;

  /** */
  int MaxHeapSize;

  /** Constructeur de la classe Tas.
      @param cartag accs  l'ensemble des informations du systme. 
  */
  Tas();

  /** Destructeur de la classe Tas. */
  ~Tas();

  /** Initialisation du Tas. 
    @param cartag accs  l'ensemble des informations du systme. 
    @param mhs taille max du tas
  */
  void Init(CartaGenePtr cartag, int mhs = MAXHEAPSIZE);

  /** Initialisation du Tas.
      Le chrono ainsi que le nombre d'appels  EM ne sont pas remis  0.
      @param cartag accs  l'ensemble des informations du systme. 
      @param mhs taille max du tas
  */
  void Initsoft(CartaGenePtr cartag, int mhs = MAXHEAPSIZE);
  
  /** Affiche les cartes contenues dans le Tas.
      Comparaison de l'ordre des marqueurs par rapport
       la meilleur carte
      @param  nbm le nombre de marqueur minimum par intervalles
      @param  blank indice ou blanc
      @param  comp comprsion des intervalles ou pas
  */
  void PrintO(int nbm, int blank, int comp) const;

  /** Affiche les cartes contenues dans le Tas. */
  void Print(void) const;

  /** Affiche les cartes contenues dans le Tas. */
  void PrintD(void) const;

  /** Affiche les cartes contenues dans le Tas. */
  void PrintSort() const;

  /** retourne un tableau d'indice de cartes contenues dans le Tas(tries). 
      @return le tableau d'indice.
  */
  intPtr IdSorted() const;

  /** Affiche les cartes contenues dans le Tas. */
  void PrintDSort() const;

  /** Insert un nouvel lment (une carte est copie). */
  int Insert(Carte *map, int iter);
  
  /** Supprime le plus mauvais lment du tas.
      Suppose que le tas est non vide.
  */
  int Extract(void);

  
  /** Fournit des informations "statistiques"
      
      @param thres un seuil de LOD.
      @return une chaine de car. ....
  */
  char* Stat(double thres);
  
  /** Quelle est la meilleure carte.
      @return  un pointeur sur le meiileur lment du tas.
  */
  Carte *Best(void) const;

  /** Quel est le Delta entre la meilleure et sa suivante
      @return  le delta.
  */
  double Delta(void) const;

  /** Quelle est le dernire carte dans la liste.
      Retourne un pointeur sur le plus mauvais lment du tas.
  */

  inline void Equi(int flag) {  
    if (flag < 0 || flag > 1) 
      {
	sprintf(bouf,  "Error : possible values are 0 or 1\n");
	perr(bouf);
	return;
      }
    
    EquivalenceFlag = flag;
  };

  /** Quelle est le dernire carte dans la liste.
      Retourne un pointeur sur le plus mauvais lment du tas.
  */

  inline Carte *Worst(void) { return Heap[0]->map; };

  /** retourne la carte  partir de sont Id.
      Retourne un pointeur sur le plus mauvais lment du tas.
  */

  inline Carte *MapFromId(int id) const { return Heap[id]->map; };
    
  /** Quelle est le dernire carte dans la liste.
	Retourne un pointeur sur la structure HMap contenant le plus mauvais
	lment du tas.
  */
  inline HMap *Worst2(void) { return Heap[0]; };
      
  /** Sauve les rsultats.
      Enregistre la meilleur carte dans le fichier mapfile, enregistre dans
      le fichier otherfile un certains nombre de rsultats dont : le temps
      de calcul (temps utilisateur pour systmes Unix, temps coul pour
      Windows et MacOS), le nombre de cartes ayant un LogLike suprieur au 
      seuil et le nombre d'appel aux mthodes EM.
      @param mapfile Fichier dans lequel est enregistr la meilleure carte.
      @param otherfile Fichier dans lequel est enregistr les rsultats statistiques.
      @param thres Seuil ????????
      @param quiet Drapeau du mode silencieux indiquant si les rsultats sont affichs.
      @param GT le gnotype sur lequel porte les calculs.
      @param Le rsum des cartes contenues dans le tas sous forme de liste.
    */

//  ListeMiniCarte * WriteBest(char *mapfile, char *otherfile, double thres,
//			     int quiet, Genotype *GT);
  
 private:
  /** */
  double worst;
  HMap *BestMap;
  
  /** */
  HMap *HashTable[HASHSIZE];
  
  /** */
  HMap **Heap;
  
  /** */
  Chronometre Chrono;

  /** */
  HMap **HashLocate(const Carte *map, HMap** ptr);

  /** probablement nullasse. */
  unsigned int HashMap(const Carte *map);

  /** Force les proprits du tas. */
  void Heapify(int i);

  /** Force les proprits du tas. */
  void heapify(HMap*** aheap, int heapsize, int i) const;
};

#endif /* _TAS_H */
