#include <iostream>
#include <fstream>
#include <string>
#include <limits.h>
#include <errno.h>
#include <math.h>
#include "CartaGene.h"
#include "BioJeu.h"
#include "BioJeuSingle.h"
#include "BJS_RH.h"
#include "BJS_BC.h"
#include "tsp.h"

#ifdef sun4
#include <ieeefp.h>
#endif

#ifdef __WIN32__
#include <float.h>
#endif

#ifdef LKHLICENCE
#include "LKH.h"
#endif

using namespace std; /* used to access ofstream and endl */

#define INFINI 10000000
#define MULTIPLY 1000
#define UNKNOWNCOST 0  /* multiplied by MULTIPLY */
#define EPSILONFLOAT 1e-100

/* Return the number of Obligate Chromosome Breaks */
double BioJeuSingle::ContribOCB(int m1, int m2)
{
  if (Cross != RH && Cross != BC) {
    sprintf(bouf,"Error: OCB criterion not implemented for this data type.\n");
    pout(bouf);
    throw BioJeu::NotImplemented();
    return 0.;
  }

  int ocb = 0;
  for (int m = 1; m <= TailleEchant; m++) {
    if (((GetEch(m1, m) == 1) && (GetEch(m2, m) == 0)) || 
	((GetEch(m1, m) == 0) && (GetEch(m2, m) == 1))) {
      ocb++;
    }
  }
  return (double) ocb;
}

/* Return the number of Obligate Chromosome Breaks */
double BioJeuSingle::NormContribOCB(int m1, int m2)
{
  if (Cross != RH && Cross != BC) {
    sprintf(bouf,"Error: normalized OCB criterion not implemented for this data type.\n");
    pout(bouf);
    throw BioJeu::NotImplemented();
    return 0.;
  }

  int knowns = 0;
  int ocb = 0;
  for (int m = 1; m <= TailleEchant; m++) {
    if ((GetEch(m1, m) != Obs1111) && (GetEch(m2, m) != Obs1111)) {
      knowns++;
      if (((GetEch(m1, m) == 1) && (GetEch(m2, m) == 0)) || 
	  ((GetEch(m1, m) == 0) && (GetEch(m2, m) == 1))) {
	ocb++;
      }
    }
  }
  if (knowns == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }
  return (((double) ocb) * TailleEchant) / knowns;
}

int CartaGene::OCB(int nbmap) const
{
 
  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return -1;
    }

  if ( nbmap < 0 || nbmap >= Heap.HeapSize )
    {
      sprintf(bouf,  "Error : This map does not exist.\n");
      perr(bouf);
      return -1;
    }

  Carte *carte = Heap.MapFromId(nbmap);
  int ocb = 0;
  try {
    for (int j = 0; j < NbMS-1; j++) {
      ocb += (int) ArbreJeu->ContribOCB(carte->ordre[j], carte->ordre[j+1]);
    }
  } catch (BioJeu::NotImplemented) { }
  return ocb;
}

double BJS_RH::Retention2pt()
{
  int n1 = 0;
  int n = 0;
  for (int m1 = 0; m1 < Cartage->NbMS; m1++) {
    for (int m = 1; m <= TailleEchant; m++) {
      if (GetEch(Cartage->MarkSelect[m1], m) != Obs1111) {
	if (GetEch(Cartage->MarkSelect[m1], m) == 1) {
	  n1++;
	}
	n++;
      }
    }
  }
  if (n==0) return 1.;
  else return (double)n1/n;
}

void BJS_RH::InitContribLogLike2pt()
{
  RetentionRH = Retention2pt();
}

/* Return -log10(likelihood(m1)) */
/* n1=n+, n0=n- */
double BJS_RH::ContribLogLike2pt(int m1)
{
  int ss[2] = {0,0};
  for (int m = 1; m <= TailleEchant; m++)    {
    if (GetEch(m1, m) != Obs1111) {
      ss[GetEch(m1, m)]++;
    }
  }
  int n1 = ss[1];
  int n0 = ss[0];
  double q = 1 - RetentionRH;
  
  return -(n1 * log10(RetentionRH) / 2 + n0 * log10(q) / 2);
}

double BJS_RH::NormContribLogLike2pt(int m1)
{
  int ss[2] = {0,0};
  for (int m = 1; m <= TailleEchant; m++)    {
    if (GetEch(m1, m) != Obs1111) {
      ss[GetEch(m1, m)]++;
    }
  }
  int n1 = ss[1];
  int n0 = ss[0];
  double q = 1 - RetentionRH;
  
  return -((double) TailleEchant * (n1 * log10(RetentionRH) / 2 + n0 * log10(q) / 2) / (n0 + n1));
}

/* Return -log10(likelihood(m1,m2)) */
/* n corresponds to the extended TSP reduction as defined by Agarwala00 */
/* n = number of knowns */
/* n1=n++=n11, n2=n10+n01= n+- + n-+, n3=n--=n00 */
double BJS_RH::ContribLogLike2pt(int m1, int m2)
{
  int ss[4] = {0,0,0,0};
  for (int m = 1; m <= TailleEchant; m++)    {
    if ((GetEch(m1, m) != Obs1111) && (GetEch(m2, m) != Obs1111)) {
      ss[GetEch(m1, m)*2 + GetEch(m2, m)]++;
    }
  }
  int n1 = ss[3];
  int n2 = ss[1] + ss[2];
  int n3 = ss[0];
  int n = n1+n2+n3;
  if (n == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }
  double q = 1 - RetentionRH;

  double a = RetentionRH*q*n;
  double b = n - RetentionRH*n1 - q*n3; /* = -b */
  double c = n2;
  double delta = b*b - 4*a*c;

  if (delta < 0) {
    sprintf(bouf,"Warning!!! The breakage probability is unknown (negative delta=%f) for markers %d,%d.\n",delta,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  double theta = (b - sqrt(delta))/(2*a);
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in square root function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  if (theta > 1) theta = 1;
  if (theta < EM_MIN_THETA_HR) theta = EM_MIN_THETA_HR;

  double loglike = (n1*log10(1-theta*q) + n2*(log10(theta) + (log10(RetentionRH)+log10(q))/2) + n3*log10(1-theta*RetentionRH));
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in log10 function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }

#ifdef __WIN32__
  if (!_finite(loglike)) {
#else
  if (!finite(loglike)) {
#endif

    sprintf(bouf,"Warning!!! Infinite loglike for markers m1=%d, m2=%d: loglike=%f\n",m1,m2,loglike);
    pout(bouf);
  }

  return -loglike;
}

/* return theta computed analytically */
double BJS_RH::Theta2pt(int m1, int m2, double retentionRH)
{
  int ss[4] = {0,0,0,0};
  for (int m = 1; m <= TailleEchant; m++)    {
    if ((GetEch(m1, m) != Obs1111) && (GetEch(m2, m) != Obs1111)) {
      ss[GetEch(m1, m)*2 + GetEch(m2, m)]++;
    }
  }
  int n1 = ss[3];
  int n2 = ss[1] + ss[2];
  int n3 = ss[0];
  int n = n1+n2+n3;
  if (n == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }
  double q = 1 - retentionRH;

  double a = retentionRH*q*n;
  double b = n - retentionRH*n1 - q*n3; /* = -b */
  double c = n2;
  double delta = b*b - 4*a*c;

  if (delta < 0) {
    sprintf(bouf,"Warning!!! The breakage probability is unknown (negative delta=%f) for markers %d,%d.\n",delta,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  double theta = (b - sqrt(delta))/(2*a);
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in square root function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  if (theta > 1) theta = 1;
  if (theta < EM_MIN_THETA_HR) theta = EM_MIN_THETA_HR;

  return theta;
}

/* n = number of knowns */
/* n1=n++=n11, n2=n10+n01= n+- + n-+, n3=n--=n00 */
double BJS_RH::NormContribLogLike2pt(int m1, int m2)
{
  int ss[9] = {0,0,0,0,0,0,0,0,0};
  for (int m = 1; m <= TailleEchant; m++)    {
    int ech1 = ((GetEch(m1, m) != Obs1111) ? GetEch(m1, m) : 2);
    int ech2 = ((GetEch(m2, m) != Obs1111) ? GetEch(m2, m) : 2);
    ss[ech1 * 3 + ech2]++;
  }
  int n00 = ss[0];
  int n01 = ss[1];
  int n02 = ss[2];
  int n10 = ss[3];
  int n11 = ss[4];
  int n12 = ss[5];
  int n20 = ss[6];
  int n21 = ss[7];
  int n22 = ss[8];
  int n1 = n11;
  int n2 = n10 + n01;
  int n3 = n00;
  int n = n1+n2+n3;
  if (n == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }
  double q = 1 - RetentionRH;

  double a = RetentionRH*q*n;
  double b = n - RetentionRH*n1 - q*n3; /* = -b */
  double c = n2;
  double delta = b*b - 4*a*c;

  if (delta < 0) {
    sprintf(bouf,"Warning!!! The breakage probability is unknown (negative delta=%f) for markers %d,%d.\n",delta,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  double theta = (b - sqrt(delta))/(2*a);
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in square root function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  if (theta > 1) theta = 1;
  if (theta < EM_MIN_THETA_HR) theta = EM_MIN_THETA_HR;

  double A = log10(1-theta*RetentionRH);
  double B = log10(1-theta*q);
  double C = (log10(theta) + (log10(RetentionRH)+log10(q))/2);
  double loglike = ((double) TailleEchant * (n00*A + n11*B + (n10+n01)*C) / n);
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in log10 function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }

#ifdef __WIN32__
  if (!_finite(loglike)) {
#else
  if (!finite(loglike)) {
#endif
    sprintf(bouf,"Warning!!! Infinite loglike for markers m1=%d, m2=%d: loglike=%f\n",m1,m2,loglike);
    pout(bouf);
  }

  return -loglike;
}

double BJS_BC::ContribLogLike2pt(int m1)
{
  return -log10(0.5) * TailleEchant / 2.;
}

double BJS_BC::ContribLogLike2pt(int m1, int m2)
{
  int nrec = 0; 
  int ndata = 0;
  
  for (int i = 1; i <= TailleEchant; i++)
    if ((GetEch(m1, i) != Obs1111) && (GetEch(m2, i) != Obs1111))
      {
	ndata++;
	if (GetEch(m1, i) != GetEch(m2, i)) nrec++;
      }

  if (ndata == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }

  double theta = (double) nrec/ndata;
  if (theta > EM_MAX_THETA) theta = EM_MAX_THETA;
  if (theta < EM_MIN_THETA) theta = EM_MIN_THETA;

  double loglike = (ndata-nrec)*log10(1-theta) + nrec*log10(theta);
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in log10 function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  return -loglike;
}

double BJS_BC::NormContribLogLike2pt(int m1, int m2)
{
  int nrec = 0; 
  int ndata = 0;
  
  for (int i = 1; i <= TailleEchant; i++)
    if ((GetEch(m1, i) != Obs1111) && (GetEch(m2, i) != Obs1111))
      {
	ndata++;
	if (GetEch(m1, i) != GetEch(m2, i)) nrec++;
      }

  if (ndata == 0) {
    //    printf("Warning!!! No known for markers %d,%d.\n",m1,m2); 
    return UNKNOWNCOST; /* valeur arbitraire! */
  }

  double theta = (double) nrec/ndata;
  if (theta > EM_MAX_THETA) theta = EM_MAX_THETA;
  if (theta < EM_MIN_THETA) theta = EM_MIN_THETA;

  double loglike = (double) TailleEchant * ((ndata-nrec)*log10(1-theta) + nrec*log10(theta)) / ndata;
  if (errno == EDOM || errno == ERANGE) {
    sprintf(bouf,"Warning!!! Error in log10 function (errno=%d, EDOM=%d) for markers %d,%d.\n",errno,EDOM,m1,m2);
    pout(bouf);
    return HUGE_VAL; /* valeur arbitraire! */
  }
  return -loglike;
}

double contribLogLike2pt1(BioJeu *ArbreJeu, int m1) {return ArbreJeu->ContribLogLike2pt(m1);}
double contribLogLike2pt2(BioJeu *ArbreJeu, int m1, int m2) {return ArbreJeu->ContribLogLike2pt(m1,m2);}
double normContribLogLike2pt1(BioJeu *ArbreJeu, int m1) {return ArbreJeu->NormContribLogLike2pt(m1);}
double normContribLogLike2pt2(BioJeu *ArbreJeu, int m1, int m2) {return ArbreJeu->NormContribLogLike2pt(m1,m2);}
double contribOCB(BioJeu *ArbreJeu, int m1, int m2) {return ArbreJeu->ContribOCB(m1,m2);}
double normContribOCB(BioJeu *ArbreJeu, int m1, int m2) {return ArbreJeu->NormContribOCB(m1,m2);}
double contribZero(BioJeu *ArbreJeu, int m1) {return 0.;}
double contribLOD(BioJeu *ArbreJeu, int m1, int m2) {return -1. * ArbreJeu->GetTwoPointsLOD(m1,m2);}
double contribHaldane(BioJeu *ArbreJeu, int m1, int m2) {return ArbreJeu->GetTwoPointsDH(m1,m2);}

void cg2tsp(BioJeu *ArbreJeu, int NbMS, int MarkSelect[], char *tspName, funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{
  ofstream tsp;
  ofstream & Tsp = tsp;
  Tsp.open(tspName);
  Tsp << "NAME: " << NbMS << "_" << MULTIPLY << ".cg" << endl;
  Tsp << "TYPE: TSP" << endl;
  Tsp << "DIMENSION: " << NbMS+1 << endl;
  Tsp << "EDGE_WEIGHT_TYPE: EXPLICIT" << endl;
  Tsp << "EDGE_WEIGHT_FORMAT: UPPER_ROW" << endl;
  Tsp << "EDGE_WEIGHT_SECTION" << endl;
  try {
    for (int m1 = 0; m1 < NbMS; m1++) {
      for (int m2 = m1+1; m2 <= NbMS; m2++) {
	double res = 0;
	if (m1 == m2) {
	  res = 0;
	} else {
	  if (m1 == 0 || m2 == 0) {
	    if (m1 == 0) {
	      res = MULTIPLY * mycontribLogLike2pt1(ArbreJeu, MarkSelect[m2-1]);
	    } else {
	      res = MULTIPLY * mycontribLogLike2pt1(ArbreJeu, MarkSelect[m1-1]);
	    }	    
	  } else {
	    res = mycontribLogLike2pt2(ArbreJeu, MarkSelect[m1-1],MarkSelect[m2-1]);
	    if (res != HUGE_VAL) {
	      res *= MULTIPLY;
	    }
	  }
	}

#ifdef __WIN32__
  if (!(_finite(res)) || res == HUGE_VAL || res > INT_MAX) { /* res > INFINI */
#else
  if (!(finite(res)) || res == HUGE_VAL || res > INT_MAX) { /* res > INFINI */
#endif
	  Tsp << INFINI << " ";
	} else {
	  Tsp << (int)res << " ";
	}
      }
      Tsp << endl;
    }
  } catch (BioJeu::NotImplemented) { }
  Tsp.close();
}

/* Create a TSPLIB format file, whose name is tspName, with the current loaded data */
/* 0 is the dummy marker (start city) */
/* note that TSPLIB city numbers starts from 1 and not 0 */
/* so, the start city is number 1 */
void CartaGene::cg2tsp(char *tspName)
{
  if (ArbreJeu == NULL) 
    {
      sprintf(bouf,  "Error: No Data Set Loaded..\n");
      perr(bouf);
      return;
    }

  /* initialisation des structures utiles aux fonctions ContribLogLike2pt */
  ArbreJeu->InitContribLogLike2pt(); 
  
  sprintf(bouf,"All TSP costs multiplied by %d.\n", MULTIPLY);
  pout(bouf);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspName, contribLogLike2pt1, contribLogLike2pt2);
  char tspnameext[256];
  strcpy(tspnameext, "norm");
  strcat(tspnameext, tspName);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspnameext, normContribLogLike2pt1, normContribLogLike2pt2);
  strcpy(tspnameext, "lod");
  strcat(tspnameext, tspName);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspnameext, contribZero, contribLOD);
  strcpy(tspnameext, "dist");
  strcat(tspnameext, tspName);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspnameext, contribZero, contribHaldane);
  strcpy(tspnameext, "ocb");
  strcat(tspnameext, tspName);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspnameext, contribZero, contribOCB);
  strcpy(tspnameext, "ocbnorm");
  strcat(tspnameext, tspName);
  ::cg2tsp(ArbreJeu, NbMS, MarkSelect, tspnameext, contribZero, normContribOCB);
}

#ifdef LKHLICENCE
Problem * cg2lkh(BioJeu *ArbreJeu, int NbMS, int MarkSelect[], funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{
  Problem *p = DefaultProblem();
  p->NAME = (char *) calloc(256, sizeof(char));
  sprintf(p->NAME, "%d_%d.cg", NbMS, MULTIPLY);
  p->TYPE = "TSP";
  p->DIMENSION = NbMS+1;
  p->EDGE_WEIGHT_TYPE = "EXPLICIT";
  // p->EDGE_WEIGHT_FORMAT = "UPPER_ROW";
  // p->C = (long *) calloc((NbMS+1) * NbMS / 2, sizeof(long));
  p->EDGE_WEIGHT_FORMAT = "FULL_MATRIX"; /* more simple to manage in iterlkh */
  p->C = (long *) calloc((NbMS+1) * (NbMS+1), sizeof(long));
  int index = 0;
  try {
    // for (int m1 = 0; m1 < NbMS; m1++) {
    //   for (int m2 = m1+1; m2 <= NbMS; m2++) {
    for (int m1 = 0; m1 <= NbMS; m1++) {
      for (int m2 = 0; m2 <= NbMS; m2++) {
	double res = 0;
	if (m1 == m2) {
	  res = 0;
	} else {
	  if (m1 == 0 || m2 == 0) {
	    if (m1 == 0) {
	      res = MULTIPLY * mycontribLogLike2pt1(ArbreJeu, MarkSelect[m2-1]);
	    } else {
	      res = MULTIPLY * mycontribLogLike2pt1(ArbreJeu, MarkSelect[m1-1]);
	    }	    
	  } else {
	    res = mycontribLogLike2pt2(ArbreJeu, MarkSelect[m1-1],MarkSelect[m2-1]);
	    if (res != HUGE_VAL) {
	      res *= MULTIPLY;
	    }
	  }
	}
#ifdef __WIN32__
  if (!(_finite(res)) || res == HUGE_VAL || res > INT_MAX) { /* res > INFINI */
#else
  if (!(finite(res)) || res == HUGE_VAL || res > INT_MAX) { /* res > INFINI */
#endif
  	  p->C[index++] = INFINI;
	} else {
	  p->C[index++] = (long)res;
	}
      }
    }
  } catch (BioJeu::NotImplemented) { FreeProblem(p); p = NULL; }
  return p;
}

CartaGene *CurrentCartage = NULL;
double BestTSPMapCost = -1e100;
int AlwaysComputeEM = 0;
int *CurrentStopFlag = NULL;
Carte *LastTSPMap = NULL;

void tsp2cg(int NbMS, long *Tour, int verbose)
{
  int newMarkSelect[NbMS];
  newMarkSelect[0] = 0;
  int dummymrk = 0;
  for (int i = 1; i <= NbMS + 1; i++) {
    if (Tour[i] == 1) {
      dummymrk = i;
      break;
    }
  }
  if (dummymrk == 0) {
    sprintf(bouf,"error, dummymarker not found!!!");
    perr(bouf);
    exit(2);
  }
  int index = 0;
  for (int i = dummymrk + 1; i <= NbMS + 1; i++) {
    newMarkSelect[index++] = CurrentCartage->MarkSelect[Tour[i] - 2];
  }
  for (int i = 1; i <= dummymrk - 1; i++) {
    newMarkSelect[index++] = CurrentCartage->MarkSelect[Tour[i] - 2];
  }

  Carte *TheMap = new Carte(CurrentCartage, NbMS, newMarkSelect);
  if (LastTSPMap != NULL) delete LastTSPMap;
  LastTSPMap = TheMap;
  
  double mycost = CurrentCartage->ComputeEM(TheMap);
  if (!CurrentCartage->QuietFlag && 
      mycost > ((CurrentCartage->Heap.HeapSize)?CurrentCartage->Heap.Best()->coutEM:-1e100) + 0.001) {
    sprintf(bouf,"[%.2f]\n",mycost);
    pout(bouf);
  }
  CurrentCartage->Heap.Insert(TheMap,0);
  if (mycost > BestTSPMapCost) BestTSPMapCost = mycost;

  if (verbose && !CurrentCartage->QuietFlag) {
    if (CurrentCartage->VerboseFlag>1) CurrentCartage->PrintDMap(TheMap, 0, TheMap);
    else CurrentCartage->PrintMap(TheMap);  
  }
}
void tsp2cg(int NbMS, long *Tour) {tsp2cg(NbMS, Tour, 0);}

void CartaGene::lkh(int nbrun, int backtrack, funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }

  /* initialisation des structures utiles aux fonctions ContribLogLike2pt */
  ArbreJeu->InitContribLogLike2pt(); 
  CurrentCartage = this;
  BestTSPMapCost = -1e100;
  AlwaysComputeEM = ((backtrack >= 0)?1:0);
  CurrentStopFlag = &(CurrentCartage->StopFlag);
  if (LastTSPMap != NULL) {
    delete LastTSPMap;
    LastTSPMap = NULL;
  }
    
  Chronometre cpu = Chronometre();
  cpu.Init();

  Problem *p = cg2lkh(ArbreJeu, NbMS, MarkSelect, mycontribLogLike2pt1, mycontribLogLike2pt2);
  if (p != NULL) {
    p->TRACE_LEVEL = -1;
    p->RUNS = nbrun;
    p->BACKTRACK_MOVE_TYPE = ((backtrack >= 0)?backtrack:0);
    // p->PRECISION = 1;
    LKH(p);

    if (!CurrentCartage->QuietFlag) {
      sprintf(bouf,"Best map with log10-likelihood = %.2f\n", BestTSPMapCost);
      pout(bouf);
      if (!StopFlag) {
	sprintf(bouf,"TSP: optimum= %f lowerbound= %f gap= %f%% totaltime= %.2f\n", p->BestCost / MULTIPLY, p->LowerBound / MULTIPLY, (p->BestCost - p->LowerBound) / p->LowerBound * 100., cpu.Read());
	pout(bouf);
      }
    }
    
    if (!StopFlag) tsp2cg(NbMS, p->BestTour, 1);

    FreeProblem(p);
  }

  if (LastTSPMap != NULL) {
    delete LastTSPMap;
    LastTSPMap = NULL;
  }
}

void CartaGene::lkhiter(int nbrun, int collectmaps, double threshold, int cost, funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error: Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  if (ArbreJeu->Cross != RH && ArbreJeu->Cross != BC) {
    sprintf(bouf,"Error: Not implemented for this data type.\n");
    perr(bouf);
    return;
  }

  /* initialisation des structures utiles aux fonctions ContribLogLike2pt */
  ArbreJeu->InitContribLogLike2pt(); 
  CurrentCartage = this;
  BestTSPMapCost = -1e100;
  AlwaysComputeEM = collectmaps;
  CurrentStopFlag = &(CurrentCartage->StopFlag);
  if (LastTSPMap != NULL) {
    delete LastTSPMap;
    LastTSPMap = NULL;
  }

  Chronometre cpu = Chronometre();
  cpu.Init();
    
  Problem *p = cg2lkh(ArbreJeu, NbMS, MarkSelect, mycontribLogLike2pt1, mycontribLogLike2pt2);
  if (p != NULL) {
    p->TRACE_LEVEL = -1;
    p->RUNS = 1;
    p->BACKTRACK_MOVE_TYPE = 0;
    // p->PRECISION = 1;

    int inverseMarkSelect[MAXNBMARKERS];
    for (int i = 0; i < NbMS; i++) {
      inverseMarkSelect[MarkSelect[i]] = i + 1;
    }

    double retentionRH = ((ArbreJeu->Cross==RH)?((BJS_RH *) ArbreJeu)->Retention2pt():0.);

    for (int i = 1; i <= nbrun; i++) {
      sprintf(bouf,"Run %d:\n", i);
      pout(bouf);
      LKH(p);

      if (!CurrentCartage->QuietFlag) {
	sprintf(bouf,"Best map with log10-likelihood = %.2f\n", BestTSPMapCost);
	pout(bouf);
	if (!StopFlag) {
	  sprintf(bouf,"TSP: optimum= %f lowerbound= %f gap= %f%% totaltime= %.2f\n", p->BestCost / MULTIPLY, p->LowerBound / MULTIPLY, (p->BestCost - p->LowerBound) / p->LowerBound * 100., cpu.Read());
	  pout(bouf);
	}
      }
    
      if (StopFlag) break;
      
      tsp2cg(NbMS, p->BestTour, 1);

      int nopenalty = 1;
      double maxdiff = 0.;
      double theta2pt[NbMS];
      for (int j = 0; j < NbMS-1; j++) {
	theta2pt[j] = ((ArbreJeu->Cross==RH)?((BJS_RH *) ArbreJeu)->Theta2pt(LastTSPMap->ordre[j],LastTSPMap->ordre[j+1],retentionRH):GetTwoPointsFR(LastTSPMap->ordre[j],LastTSPMap->ordre[j+1]));
	if (LastTSPMap->tr[j] - theta2pt[j] > maxdiff) {
	  maxdiff = LastTSPMap->tr[j] - theta2pt[j];
	  nopenalty = 0;
	}
      }
      if (nopenalty) break;
      for (int j = 0; j < NbMS-1; j++) {
	if (maxdiff - (LastTSPMap->tr[j] - theta2pt[j]) <= threshold) {
	  sprintf(bouf,"%d %d %d increased by %d\n", LastTSPMap->ordre[j], LastTSPMap->ordre[j+1], p->C[inverseMarkSelect[LastTSPMap->ordre[j]] * (NbMS+1) + inverseMarkSelect[LastTSPMap->ordre[j+1]]], cost);
	  pout(bouf);
	  p->C[inverseMarkSelect[LastTSPMap->ordre[j]] * (NbMS+1) + inverseMarkSelect[LastTSPMap->ordre[j+1]]] += cost;
	  p->C[inverseMarkSelect[LastTSPMap->ordre[j+1]] * (NbMS+1) + inverseMarkSelect[LastTSPMap->ordre[j]]] += cost;
	}
      }

      // reuse best tour as next initial tour
      if (p->INITIAL_TOUR) {
	free(p->INITIAL_TOUR);
	p->INITIAL_TOUR = NULL;
      }
      p->INITIAL_TOUR = (long *) calloc(NbMS + 1, sizeof(long));
      for (int k=0; k <= NbMS; k++) {
	p->INITIAL_TOUR[k] = p->BestTour[k + 1];
      }
      free(p->BestTour);
      p->BestTour = NULL;
    }
    FreeProblem(p);
  }

  if (LastTSPMap != NULL) {
    delete LastTSPMap;
    LastTSPMap = NULL;
  }
}
#else
void CartaGene::lkh(int nbrun, int backtrack, funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{
  sprintf(bouf,"This version of CarthaGene has been compiled without Keld Helsgaun's LKH TSP software. If you want to use this command you must check that you are entitled to use this code (only for research use, see the README file in the LKH directory) and use the research version of CarthaGne(compiled with the LKHLICENCE macro defined).\n");
  pout(bouf);
}
void CartaGene::lkhiter(int nbrun, int collectmaps, double threshold, int cost, funcloglike1 mycontribLogLike2pt1, funcloglike2 mycontribLogLike2pt2)
{
  sprintf(bouf,"This version of CarthaGene has been compiled without Keld Helsgaun's LKH TSP software. If you want to use this command you must check that you are entitled to use this code (only for research use, see the README file in the LKH directory) and use the research version of CarthaGne(compiled with the LKHLICENCE macro defined).\n");
  pout(bouf);
}
#endif
